@@ -23,12 +23,12 @@ mca_mtl_ofi_module_t ompi_mtl_ofi = {
         ompi_mtl_ofi_del_procs,
         ompi_mtl_ofi_finalize,
 
-        ompi_mtl_ofi_send,
-        ompi_mtl_ofi_isend,
-        ompi_mtl_ofi_irecv,
-        ompi_mtl_ofi_iprobe,
-        ompi_mtl_ofi_imrecv,
-        ompi_mtl_ofi_improbe,
+         NULL,
+         NULL,
+         NULL,
+         NULL,
+         ompi_mtl_ofi_imrecv,
+         NULL,
 
         ompi_mtl_ofi_cancel,
         ompi_mtl_ofi_add_comm,
@@ -51,17 +51,20 @@ ompi_mtl_ofi_add_procs(struct mca_mtl_base_module_t *mtl,
     size_t namelen;
     int count = 0;
     char *ep_name = NULL;
-    char *ep_names = NULL;
     fi_addr_t *fi_addrs = NULL;
     mca_mtl_ofi_endpoint_t *endpoint = NULL;
+    int num_peers_limit = (1 << ompi_mtl_ofi.num_bits_source_rank) - 1;
 
     namelen = ompi_mtl_ofi.epnamelen;
 
-    /**
-     * Create array of EP names.
-     */
-    ep_names = malloc(nprocs * namelen);
-    if (NULL == ep_names) {
+    /* We cannot add more ranks than available tag bits */
+    if ((false == ompi_mtl_ofi.fi_cq_data) &&
+        OPAL_UNLIKELY(((int) (nprocs + ompi_mtl_ofi.num_peers) > num_peers_limit))) {
+        opal_output(0, "%s:%d: OFI provider: %s does not have enough bits for source rank in its tag.\n"
+                       "Adding more ranks will result in undefined behaviour. Please enable\n"
+                       "FI_REMOTE_CQ_DATA feature in the provider. For more info refer fi_cq(3).\n",
+                       __FILE__, __LINE__, ompi_mtl_ofi.provider_name);
+        fflush(stderr);
         ret = OMPI_ERROR;
         goto bail;
     }
@@ -75,34 +78,34 @@ ompi_mtl_ofi_add_procs(struct mca_mtl_base_module_t *mtl,
         goto bail;
     }
 
-    /**
-     * Retrieve the processes' EP names from modex.
-     */
     for (i = 0; i < nprocs; ++i) {
+        /**
+         * Retrieve the processes' EP name from modex.
+         */
         OFI_COMPAT_MODEX_RECV(ret,
                               &mca_mtl_ofi_component.super.mtl_version,
                               procs[i],
                               (void**)&ep_name,
                               &size);
         if (OMPI_SUCCESS != ret) {
-            opal_output_verbose(1, ompi_mtl_base_framework.framework_output,
-                                "%s:%d: modex_recv failed: %d\n",
-                                __FILE__, __LINE__, ret);
+            opal_show_help("help-mtl-ofi.txt", "modex failed",
+                           true, ompi_process_info.nodename,
+			   procs[i]->super.proc_hostname,
+			   opal_strerror(ret), ret);
             goto bail;
         }
-        memcpy(&ep_names[i*namelen], ep_name, namelen);
-    }
 
-    /**
-     * Map the EP names to fi_addrs.
-     */
-    count = fi_av_insert(ompi_mtl_ofi.av, ep_names, nprocs, fi_addrs, 0, NULL);
-    if ((count < 0) || (nprocs != (size_t)count)) {
-        opal_output_verbose(1, ompi_mtl_base_framework.framework_output,
-                            "%s:%d: fi_av_insert failed: %d\n",
-                            __FILE__, __LINE__, count);
-        ret = OMPI_ERROR;
-        goto bail;
+        /**
+         * Map the EP name to fi_addr.
+         */
+        count = fi_av_insert(ompi_mtl_ofi.av, ep_name, 1, &fi_addrs[i], 0, NULL);
+        if ((count < 0) || (1 != (size_t)count)) {
+            opal_output_verbose(1, opal_common_ofi.output,
+                                "%s:%d: fi_av_insert failed for address %s: %d\n",
+                                __FILE__, __LINE__, ep_name, count);
+            ret = OMPI_ERROR;
+            goto bail;
+        }
     }
 
     /**
@@ -111,7 +114,7 @@ ompi_mtl_ofi_add_procs(struct mca_mtl_base_module_t *mtl,
     for (i = 0; i < nprocs; ++i) {
         endpoint = OBJ_NEW(mca_mtl_ofi_endpoint_t);
         if (NULL == endpoint) {
-            opal_output_verbose(1, ompi_mtl_base_framework.framework_output,
+            opal_output_verbose(1, opal_common_ofi.output,
                                 "%s:%d: mtl/ofi: could not allocate endpoint"
                                 " structure\n",
                                 __FILE__, __LINE__);
@@ -126,15 +129,15 @@ ompi_mtl_ofi_add_procs(struct mca_mtl_base_module_t *mtl,
         procs[i]->proc_endpoints[OMPI_PROC_ENDPOINT_TAG_MTL] = endpoint;
     }
 
+    /* Update global counter of number of procs added to this rank */
+    ompi_mtl_ofi.num_peers += nprocs;
+
     ret = OMPI_SUCCESS;
 
 bail:
     if (fi_addrs)
         free(fi_addrs);
 
-    if (ep_names)
-        free(ep_names);
-
     return ret;
 }
 
@@ -153,7 +156,7 @@ ompi_mtl_ofi_del_procs(struct mca_mtl_base_module_t *mtl,
             endpoint = procs[i]->proc_endpoints[OMPI_PROC_ENDPOINT_TAG_MTL];
             ret = fi_av_remove(ompi_mtl_ofi.av, &endpoint->peer_fiaddr, 1, 0);
             if (ret) {
-                opal_output_verbose(1, ompi_mtl_base_framework.framework_output,
+                opal_output_verbose(1, opal_common_ofi.output,
                         "%s:%d: fi_av_remove failed: %s\n", __FILE__, __LINE__, fi_strerror(errno));
                 return ret;
             }
