@@ -1,3 +1,4 @@
+/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil -*- */
 /*
  * Copyright (c) 2004-2005 The Trustees of Indiana University and Indiana
  *                         University Research and Technology
@@ -5,10 +6,15 @@
  * Copyright (c) 2004-2013 The University of Tennessee and The University
  *                         of Tennessee Research Foundation.  All rights
  *                         reserved.
- * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart, 
+ * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart,
  *                         University of Stuttgart.  All rights reserved.
  * Copyright (c) 2004-2006 The Regents of the University of California.
  *                         All rights reserved.
+ * Copyright (c) 2015-2017 Los Alamos National Security, LLC. All rights
+ *                         reserved.
+ * Copyright (c) 2015      Research Organization for Information Science
+ *                         and Technology (RIST). All rights reserved.
+ * Copyright (c) 2017      Intel, Inc. All rights reserved
  * $COPYRIGHT$
  *
  * Additional copyrights may follow
@@ -26,7 +32,7 @@
 #include "ompi/mca/mtl/mtl.h"
 #include "opal/prefetch.h"
 
-struct mca_pml_cm_send_request_t { 
+struct mca_pml_cm_send_request_t {
     mca_pml_cm_request_t req_base;
     mca_pml_base_send_mode_t req_send_mode;
 };
@@ -34,7 +40,7 @@ typedef struct mca_pml_cm_send_request_t mca_pml_cm_send_request_t;
 OBJ_CLASS_DECLARATION(mca_pml_cm_send_request_t);
 
 
-struct mca_pml_cm_thin_send_request_t { 
+struct mca_pml_cm_thin_send_request_t {
     mca_pml_cm_send_request_t req_send;
     mca_mtl_request_t req_mtl;            /**< the mtl specific memory. This field should be the last in the struct */
 };
@@ -44,7 +50,7 @@ OBJ_CLASS_DECLARATION(mca_pml_cm_thin_send_request_t);
 
 struct mca_pml_cm_hvy_send_request_t {
     mca_pml_cm_send_request_t req_send;
-    void *req_addr;                       /**< pointer to application buffer */
+    const void *req_addr;                 /**< pointer to application buffer */
     size_t req_count;                     /**< count of user datatype elements */
     int32_t req_peer;                     /**< peer process - rank w/in this communicator */
     int32_t req_tag;                      /**< user defined tag */
@@ -60,14 +66,13 @@ OBJ_CLASS_DECLARATION(mca_pml_cm_hvy_send_request_t);
 #define MCA_PML_CM_THIN_SEND_REQUEST_ALLOC(sendreq, comm, dst,          \
                                            ompi_proc)                   \
 do {                                                                    \
-    ompi_free_list_item_t* item;                                        \
     ompi_proc = ompi_comm_peer_lookup( comm, dst );                     \
                                                                         \
     if(OPAL_UNLIKELY(NULL == ompi_proc)) {                              \
         sendreq = NULL;                                                 \
     } else {                                                            \
-        OMPI_FREE_LIST_WAIT_MT(&mca_pml_base_send_requests, item);         \
-        sendreq = (mca_pml_cm_thin_send_request_t*)item;                \
+        sendreq = (mca_pml_cm_thin_send_request_t*)                     \
+          opal_free_list_wait (&mca_pml_base_send_requests);            \
         sendreq->req_send.req_base.req_pml_type = MCA_PML_CM_REQUEST_SEND_THIN; \
         sendreq->req_mtl.ompi_req = (ompi_request_t*) sendreq;          \
         sendreq->req_mtl.completion_callback = mca_pml_cm_send_request_completion; \
@@ -77,9 +82,8 @@ do {                                                                    \
 #define MCA_PML_CM_THIN_SEND_REQUEST_ALLOC(sendreq, comm, dst,          \
                                            ompi_proc)                   \
 do {                                                                    \
-    ompi_free_list_item_t* item;                                        \
-    OMPI_FREE_LIST_WAIT_MT(&mca_pml_base_send_requests, item);		\
-    sendreq = (mca_pml_cm_thin_send_request_t*)item;			\
+    sendreq = (mca_pml_cm_thin_send_request_t*)                         \
+        opal_free_list_wait (&mca_pml_base_send_requests);              \
     sendreq->req_send.req_base.req_pml_type = MCA_PML_CM_REQUEST_SEND_THIN; \
     sendreq->req_mtl.ompi_req = (ompi_request_t*) sendreq;              \
     sendreq->req_mtl.completion_callback = mca_pml_cm_send_request_completion; \
@@ -91,13 +95,12 @@ do {                                                                    \
 #define MCA_PML_CM_HVY_SEND_REQUEST_ALLOC(sendreq, comm, dst,           \
                                           ompi_proc)                    \
 {                                                                       \
-    ompi_free_list_item_t* item;                                        \
     ompi_proc = ompi_comm_peer_lookup( comm, dst );                     \
     if(OPAL_UNLIKELY(NULL == ompi_proc)) {                              \
         sendreq = NULL;                                                 \
     } else {                                                            \
-        OMPI_FREE_LIST_WAIT_MT(&mca_pml_base_send_requests, item);         \
-        sendreq = (mca_pml_cm_hvy_send_request_t*)item;                 \
+        sendreq = (mca_pml_cm_hvy_send_request_t*)                      \
+          opal_free_list_wait (&mca_pml_base_send_requests);            \
         sendreq->req_send.req_base.req_pml_type = MCA_PML_CM_REQUEST_SEND_HEAVY; \
         sendreq->req_mtl.ompi_req = (ompi_request_t*) sendreq;          \
         sendreq->req_mtl.completion_callback = mca_pml_cm_send_request_completion; \
@@ -107,9 +110,8 @@ do {                                                                    \
 #define MCA_PML_CM_HVY_SEND_REQUEST_ALLOC(sendreq, comm, dst,           \
                                           ompi_proc)                    \
 {                                                                       \
-    ompi_free_list_item_t* item;                                        \
-    OMPI_FREE_LIST_WAIT_MT(&mca_pml_base_send_requests, item);		\
-    sendreq = (mca_pml_cm_hvy_send_request_t*) item;			\
+    sendreq = (mca_pml_cm_hvy_send_request_t*)                          \
+        opal_free_list_wait (&mca_pml_base_send_requests);              \
     sendreq->req_send.req_base.req_pml_type = MCA_PML_CM_REQUEST_SEND_HEAVY; \
     sendreq->req_mtl.ompi_req = (ompi_request_t*) sendreq;              \
     sendreq->req_mtl.completion_callback = mca_pml_cm_send_request_completion; \
@@ -124,18 +126,20 @@ do {                                                                    \
                                             datatype,                   \
                                             sendmode,                   \
                                             buf,                        \
-                                            count)                      \
+                                            count,                      \
+                                            flags )                     \
 {                                                                       \
     OBJ_RETAIN(comm);                                                   \
-    OBJ_RETAIN(datatype);                                               \
+    OMPI_DATATYPE_RETAIN(datatype);                                     \
     (req_send)->req_base.req_comm = comm;                               \
     (req_send)->req_base.req_datatype = datatype;                       \
+    MCA_PML_CM_SWITCH_CUDA_CONVERTOR_OFF(flags, datatype, count);       \
     opal_convertor_copy_and_prepare_for_send(                           \
-                                             ompi_proc->proc_convertor, \
+                                             ompi_proc->super.proc_convertor, \
                                              &(datatype->super),        \
                                              count,                     \
                                              buf,                       \
-                                             0,                         \
+                                             flags,                     \
                                              &(req_send)->req_base.req_convertor ); \
     (req_send)->req_base.req_ompi.req_mpi_object.comm = comm;           \
     (req_send)->req_base.req_ompi.req_status.MPI_SOURCE =               \
@@ -153,18 +157,20 @@ do {                                                                    \
                                             datatype,                   \
                                             sendmode,                   \
                                             buf,                        \
-                                            count)                      \
+                                            count,                      \
+                                            flags )                     \
 {                                                                       \
     OBJ_RETAIN(comm);                                                   \
-    OBJ_RETAIN(datatype);                                               \
+    OMPI_DATATYPE_RETAIN(datatype);                                     \
     (req_send)->req_base.req_comm = comm;                               \
     (req_send)->req_base.req_datatype = datatype;                       \
+    MCA_PML_CM_SWITCH_CUDA_CONVERTOR_OFF(flags, datatype, count);       \
     opal_convertor_copy_and_prepare_for_send(                           \
         ompi_mpi_local_convertor,                                       \
         &(datatype->super),                                             \
         count,                                                          \
         buf,                                                            \
-        0,                                                              \
+        flags,                                                          \
         &(req_send)->req_base.req_convertor );                          \
     (req_send)->req_base.req_ompi.req_mpi_object.comm = comm;           \
     (req_send)->req_base.req_ompi.req_status.MPI_SOURCE =               \
@@ -184,18 +190,20 @@ do {                                                                    \
                                             datatype,                   \
                                             sendmode,                   \
                                             buf,                        \
-                                            count)                      \
+                                            count,                      \
+                                            flags )                     \
 {                                                                       \
     OBJ_RETAIN(comm);                                                   \
-    OBJ_RETAIN(datatype);                                               \
+    OMPI_DATATYPE_RETAIN(datatype);                                     \
     (req_send)->req_base.req_comm = comm;                               \
     (req_send)->req_base.req_datatype = datatype;                       \
+    MCA_PML_CM_SWITCH_CUDA_CONVERTOR_OFF(flags, datatype, count);       \
     opal_convertor_copy_and_prepare_for_send(                           \
-                                             ompi_proc->proc_convertor, \
+                                             ompi_proc->super.proc_convertor, \
                                              &(datatype->super),        \
                                              count,                     \
                                              buf,                       \
-                                             0,                         \
+                                             flags,                     \
                                              &(req_send)->req_base.req_convertor ); \
     (req_send)->req_base.req_ompi.req_mpi_object.comm = comm;           \
     (req_send)->req_base.req_ompi.req_status.MPI_SOURCE =               \
@@ -214,10 +222,11 @@ do {                                                                    \
                                             datatype,                   \
                                             sendmode,                   \
                                             buf,                        \
-                                            count)                      \
+                                            count,                      \
+                                            flags )                     \
 {                                                                       \
     OBJ_RETAIN(comm);                                                   \
-    OBJ_RETAIN(datatype);                                               \
+    OMPI_DATATYPE_RETAIN(datatype);                                     \
     (req_send)->req_base.req_comm = comm;                               \
     (req_send)->req_base.req_datatype = datatype;                       \
     if (opal_datatype_is_contiguous_memory_layout(&datatype->super, count)) { \
@@ -227,19 +236,22 @@ do {                                                                    \
             ompi_mpi_local_convertor->flags;                            \
         (req_send)->req_base.req_convertor.master     =                 \
             ompi_mpi_local_convertor->master;                           \
-        (req_send)->req_base.req_convertor.local_size =                 \
-            count * datatype->super.size;                               \
-        (req_send)->req_base.req_convertor.pBaseBuf   =                 \
-            (unsigned char*)buf + datatype->super.true_lb;              \
-        (req_send)->req_base.req_convertor.count      = count;          \
-        (req_send)->req_base.req_convertor.pDesc      = &datatype->super; \
+        /* Switches off device buffer  detection if                 \
+         MTL set MCA_MTL_BASE_FLAG_CUDA_INIT_DISABLE during init */     \
+        MCA_PML_CM_SWITCH_CUDA_CONVERTOR_OFF(flags, datatype, count);   \
+        (req_send)->req_base.req_convertor.flags |= flags;              \
+        /* Sets CONVERTOR_CUDA flag if device buffer */                   \
+        opal_convertor_prepare_for_send(                            \
+            &req_send->req_base.req_convertor,                      \
+            &datatype->super, count, (unsigned char*)buf); \
     } else {                                                            \
+        MCA_PML_CM_SWITCH_CUDA_CONVERTOR_OFF(flags, datatype, count);   \
         opal_convertor_copy_and_prepare_for_send(                       \
             ompi_mpi_local_convertor,                                   \
             &(datatype->super),                                         \
             count,                                                      \
             buf,                                                        \
-            0,                                                          \
+            flags,                                                      \
             &(req_send)->req_base.req_convertor );                      \
     }                                                                   \
     (req_send)->req_base.req_ompi.req_mpi_object.comm = comm;           \
@@ -262,7 +274,8 @@ do {                                                                    \
                                           persistent,                   \
                                           blocking,                     \
                                           buf,                          \
-                                          count)                        \
+                                          count,                        \
+                                          flags )                       \
     do {                                                                \
         OMPI_REQUEST_INIT(&(sendreq->req_send.req_base.req_ompi),       \
                           persistent);                                  \
@@ -277,7 +290,8 @@ do {                                                                    \
                                              datatype,                  \
                                              sendmode,                  \
                                              buf,                       \
-                                             count);                    \
+                                             count,                     \
+                                             flags )                    \
         opal_convertor_get_packed_size(                                 \
                                        &sendreq->req_send.req_base.req_convertor, \
                                        &sendreq->req_count );           \
@@ -296,7 +310,8 @@ do {                                                                    \
                                            datatype,                    \
                                            sendmode,                    \
                                            buf,                         \
-                                           count)                       \
+                                           count,                       \
+                                           flags )                      \
     do {                                                                \
         OMPI_REQUEST_INIT(&(sendreq->req_send.req_base.req_ompi),       \
                           false);                                       \
@@ -307,7 +322,8 @@ do {                                                                    \
                                              datatype,                  \
                                              sendmode,                  \
                                              buf,                       \
-                                             count);                    \
+                                             count,                     \
+                                             flags);                    \
         sendreq->req_send.req_base.req_pml_complete = false;            \
     } while(0)
 
@@ -315,7 +331,7 @@ do {                                                                    \
 #define MCA_PML_CM_SEND_REQUEST_START_SETUP(req_send)                   \
     do {                                                                \
         (req_send)->req_base.req_pml_complete = false;                  \
-        (req_send)->req_base.req_ompi.req_complete = false;             \
+        (req_send)->req_base.req_ompi.req_complete = REQUEST_PENDING;   \
         (req_send)->req_base.req_ompi.req_state =                       \
             OMPI_REQUEST_ACTIVE;                                        \
         (req_send)->req_base.req_ompi.req_status._cancelled = 0;        \
@@ -366,36 +382,39 @@ do {                                                                    \
         }                                                               \
     }                                                                   \
  } while(0);
-        
 
-#define MCA_PML_CM_HVY_SEND_REQUEST_START(sendreq, ret)                          \
-do {                                                                             \
-    ret = OMPI_SUCCESS;                                                          \
-    MCA_PML_CM_SEND_REQUEST_START_SETUP(&(sendreq)->req_send);                   \
-    if (sendreq->req_send.req_send_mode == MCA_PML_BASE_SEND_BUFFERED) {         \
-        MCA_PML_CM_HVY_SEND_REQUEST_BSEND_ALLOC(sendreq, ret);                   \
-    }                                                                            \
-    if (OMPI_SUCCESS == ret) {                                                   \
-        ret = OMPI_MTL_CALL(isend(ompi_mtl,                                      \
-                                  sendreq->req_send.req_base.req_comm,           \
-                                  sendreq->req_peer,                             \
-                                  sendreq->req_tag,                              \
-                                  &sendreq->req_send.req_base.req_convertor,     \
-                                  sendreq->req_send.req_send_mode,               \
-                                  sendreq->req_blocking,                         \
-                                  &sendreq->req_mtl));                           \
-        if(OMPI_SUCCESS == ret &&                                                \
-           sendreq->req_send.req_send_mode == MCA_PML_BASE_SEND_BUFFERED) {      \
-            sendreq->req_send.req_base.req_ompi.req_status.MPI_ERROR = 0;        \
-            ompi_request_complete(&(sendreq)->req_send.req_base.req_ompi, true); \
-        }                                                                        \
-    }                                                                            \
+
+#define MCA_PML_CM_HVY_SEND_REQUEST_START(sendreq, ret)                              \
+do {                                                                                 \
+    ret = OMPI_SUCCESS;                                                              \
+    MCA_PML_CM_SEND_REQUEST_START_SETUP(&(sendreq)->req_send);                       \
+    if (sendreq->req_send.req_send_mode == MCA_PML_BASE_SEND_BUFFERED) {             \
+        MCA_PML_CM_HVY_SEND_REQUEST_BSEND_ALLOC(sendreq, ret);                       \
+    }                                                                                \
+    if (OMPI_SUCCESS == ret) {                                                       \
+        ret = OMPI_MTL_CALL(isend(ompi_mtl,                                          \
+                                  sendreq->req_send.req_base.req_comm,               \
+                                  sendreq->req_peer,                                 \
+                                  sendreq->req_tag,                                  \
+                                  &sendreq->req_send.req_base.req_convertor,         \
+                                  sendreq->req_send.req_send_mode,                   \
+                                  sendreq->req_blocking,                             \
+                                  &sendreq->req_mtl));                               \
+        if(OMPI_SUCCESS == ret &&                                                    \
+           sendreq->req_send.req_send_mode == MCA_PML_BASE_SEND_BUFFERED) {          \
+            sendreq->req_send.req_base.req_ompi.req_status.MPI_ERROR = 0;            \
+            if(!REQUEST_COMPLETE(&sendreq->req_send.req_base.req_ompi)) {            \
+                /* request may have already been marked complete by the MTL */       \
+                ompi_request_complete(&(sendreq)->req_send.req_base.req_ompi, true); \
+            }                                                                        \
+        }                                                                            \
+    }                                                                                \
  } while (0)
 
 /*
  * The PML has completed a send request. Note that this request
  * may have been orphaned by the user or have already completed
- * at the MPI level. 
+ * at the MPI level.
  * This macro will never be called directly from the upper level, as it should
  * only be an internal call to the PML.
  */
@@ -408,9 +427,8 @@ do {
         mca_pml_base_bsend_request_free(sendreq->req_buff);                        \
     }                                                                              \
                                                                                    \
-    OPAL_THREAD_LOCK(&ompi_request_lock);                                          \
-    if( false == sendreq->req_send.req_base.req_ompi.req_complete ) {              \
-        /* Should only be called for long messages (maybe synchronous) */          \
+    if( !REQUEST_COMPLETE(&sendreq->req_send.req_base.req_ompi)) {                 \
+        /* the request may have already been marked complete by the MTL */         \
         ompi_request_complete(&(sendreq->req_send.req_base.req_ompi), true);       \
     }                                                                              \
     sendreq->req_send.req_base.req_pml_complete = true;                            \
@@ -425,7 +443,6 @@ do {
                                         &offset);                                  \
         }                                                                          \
     }                                                                              \
-    OPAL_THREAD_UNLOCK(&ompi_request_lock);                                        \
  } while (0)
 
 
@@ -435,18 +452,18 @@ do {
 #define MCA_PML_CM_HVY_SEND_REQUEST_RETURN(sendreq)                     \
     {                                                                   \
         /*  Let the base handle the reference counts */                 \
-        OBJ_RELEASE(sendreq->req_send.req_base.req_datatype);           \
+        OMPI_DATATYPE_RETAIN(sendreq->req_send.req_base.req_datatype);  \
         OBJ_RELEASE(sendreq->req_send.req_base.req_comm);               \
         OMPI_REQUEST_FINI(&sendreq->req_send.req_base.req_ompi);        \
         opal_convertor_cleanup( &(sendreq->req_send.req_base.req_convertor) ); \
-        OMPI_FREE_LIST_RETURN_MT( &mca_pml_base_send_requests,             \
-                               (ompi_free_list_item_t*)sendreq);        \
+        opal_free_list_return ( &mca_pml_base_send_requests,            \
+                               (opal_free_list_item_t*)sendreq);        \
     }
 
 /*
  * The PML has completed a send request. Note that this request
  * may have been orphaned by the user or have already completed
- * at the MPI level. 
+ * at the MPI level.
  * This macro will never be called directly from the upper level, as it should
  * only be an internal call to the PML.
  */
@@ -454,8 +471,7 @@ do {
 do {                                                                         \
     assert( false == sendreq->req_send.req_base.req_pml_complete );          \
                                                                              \
-    OPAL_THREAD_LOCK(&ompi_request_lock);                                    \
-    if( false == sendreq->req_send.req_base.req_ompi.req_complete ) {        \
+    if( !REQUEST_COMPLETE(&sendreq->req_send.req_base.req_ompi)) {           \
         /* Should only be called for long messages (maybe synchronous) */    \
         ompi_request_complete(&(sendreq->req_send.req_base.req_ompi), true); \
     }                                                                        \
@@ -464,22 +480,21 @@ do {                                                                         \
     if( sendreq->req_send.req_base.req_free_called ) {                       \
         MCA_PML_CM_THIN_SEND_REQUEST_RETURN( sendreq );                      \
     }                                                                        \
-    OPAL_THREAD_UNLOCK(&ompi_request_lock);                                  \
  } while (0)
-    
-    
+
+
 /*
  * Release resources associated with a request
  */
 #define MCA_PML_CM_THIN_SEND_REQUEST_RETURN(sendreq)                    \
     {                                                                   \
         /*  Let the base handle the reference counts */                 \
-        OBJ_RELEASE(sendreq->req_send.req_base.req_datatype);           \
+        OMPI_DATATYPE_RETAIN(sendreq->req_send.req_base.req_datatype);  \
         OBJ_RELEASE(sendreq->req_send.req_base.req_comm);               \
         OMPI_REQUEST_FINI(&sendreq->req_send.req_base.req_ompi);        \
         opal_convertor_cleanup( &(sendreq->req_send.req_base.req_convertor) ); \
-        OMPI_FREE_LIST_RETURN_MT( &mca_pml_base_send_requests,             \
-                               (ompi_free_list_item_t*)sendreq);        \
+        opal_free_list_return ( &mca_pml_base_send_requests,            \
+                               (opal_free_list_item_t*)sendreq);        \
     }
 
 extern void
