@@ -5,15 +5,18 @@
  * Copyright (c) 2004-2005 The University of Tennessee and The University
  *                         of Tennessee Research Foundation.  All rights
  *                         reserved.
- * Copyright (c) 2004-2008 High Performance Computing Center Stuttgart, 
+ * Copyright (c) 2004-2008 High Performance Computing Center Stuttgart,
  *                         University of Stuttgart.  All rights reserved.
  * Copyright (c) 2004-2005 The Regents of the University of California.
  *                         All rights reserved.
  * Copyright (c) 2006      Cisco Systems, Inc.  All rights reserved.
+ * Copyright (c) 2015-2017 Research Organization for Information Science
+ *                         and Technology (RIST). All rights reserved.
+ * Copyright (c) 2017      FUJITSU LIMITED.  All rights reserved.
  * $COPYRIGHT$
- * 
+ *
  * Additional copyrights may follow
- * 
+ *
  * $HEADER$
  */
 #include "ompi_config.h"
@@ -27,18 +30,17 @@
 #include "ompi/errhandler/errhandler.h"
 #include "ompi/memchecker.h"
 
-#if OPAL_HAVE_WEAK_SYMBOLS && OMPI_PROFILING_DEFINES
+#if OMPI_BUILD_MPI_PROFILING
+#if OPAL_HAVE_WEAK_SYMBOLS
 #pragma weak MPI_Start = PMPI_Start
 #endif
-
-#if OMPI_PROFILING_DEFINES
-#include "ompi/mpi/c/profile/defines.h"
+#define MPI_Start PMPI_Start
 #endif
 
 static const char FUNC_NAME[] = "MPI_Start";
 
 
-int MPI_Start(MPI_Request *request) 
+int MPI_Start(MPI_Request *request)
 {
     int ret = OMPI_SUCCESS;
 
@@ -59,15 +61,20 @@ int MPI_Start(MPI_Request *request)
      * MPI standard 3.1 page 78 line 19: we must have the following
      * sequence CREATE (START COMPLETE)* FREE. The upper level is
      * responsible for handling any concurency. The PML must handle
-     * this case, as it is the only one knowing if the request can 
+     * this case, as it is the only one knowing if the request can
      * be reused or not (it is PML completed or not?).
      */
 
     switch((*request)->req_type) {
     case OMPI_REQUEST_PML:
+    case OMPI_REQUEST_COLL:
+        if ( MPI_PARAM_CHECK && !((*request)->req_persistent &&
+                                  OMPI_REQUEST_INACTIVE == (*request)->req_state)) {
+            return OMPI_ERRHANDLER_INVOKE(MPI_COMM_WORLD, MPI_ERR_REQUEST, FUNC_NAME);
+        }
         OPAL_CR_ENTER_LIBRARY();
 
-        ret = MCA_PML_CALL(start(1, request));
+        ret = (*request)->req_start(1, request);
 
         OPAL_CR_EXIT_LIBRARY();
         return ret;
